-- Warborg is Wargames playing on network
-- Copyright (C) 2011 Manuel De Girardi

-- This Program is Free Software: You Can Redistribute It and/or Modify
-- It Under The Terms of The GNU General Public License As Published By
-- The Free Software Foundation, Either Version 3 of The License, or
-- (at Your Option) Any Later Version.

-- This Program is Distributed in The Hope That It Will Be Useful,
-- But WITHOUT ANY WARRANTY; Without Even The Implied Warranty of
-- MERCHANTABILITY or FITNESS for A PARTICULAR PURPOSE.  See The
-- GNU General Public License for More Details.

-- You Should Have Received A Copy of The GNU General Public License
-- Along with This Program.  if not, See <Http://Www.Gnu.Org/Licenses/>.
with Warborg.Xborg;
use Warborg, Warborg.Xborg;
with Ada.Command_Line, Ada.Text_Io;
use Ada;
with Gnat.Os_Lib, Gnat.Sockets;
use Gnat, Gnat.Sockets;
with Gtk.Main;
with Reborg_Status;
with Interfaces.C;
use Interfaces;

procedure Xborg is


   procedure Usage is

   begin
      Text_Io.Put_Line("usage : " & Command_Line.Command_Name & " Hostname [ Logname password ]");
   end Usage;

  Args_Error : exception;


  procedure Print(Text : in String; Rate : in Duration := 0.05; Quiet : in Boolean := true) is
  begin
     for I in Text'Range loop
        Text_Io.Put(Text(I));
        if not Quiet and Text(I) = Character'Val(32) then
           Text_Io.Put(Character'Val(7));
        end if;
        delay Rate;
     end loop;
  end Print;




  Term_Address      : Sock_Addr_Type;
  Socket            : Socket_Type;
  Channel           : Stream_access;
  Command, Response : String_Access;
  Status            : Boolean := False;
  Logname           : String_access;
  Reborg_Parameters : Os_Lib.Argument_List_Access;
  Reborg_Id         : Os_Lib.Process_Id;

  Xborg             : Main_Application;

begin
   if Command_Line.Argument_Count < 1 then

      raise Args_Error;
   end if;

   Term_Address.Addr := Addresses (Get_Host_By_Name (Command_Line.Argument(1)));
   Term_Address.Port := Xborg_Port;
   Create_Socket (Socket);

   Set_Socket_Option
     (Socket,
      Socket_Level,
      (Reuse_Address, True));

   delay 0.2;

   Connect_Socket (Socket, Term_Address);

   Channel :=  stream(Socket);
   String'Output(Channel, "Hello World!");
   if Command_Line.Argument_Count = 3 then

      Response := new String ' (String'Input(Channel));
      Print(Response.all, 0.02, False);
      String'Output(Channel, Command_Line.Argument(2));
      Status := Boolean'Input(Channel);
      Response := new String ' (String'Input(Channel));
      Print(Response.all, 0.02, False);
      String'Output(Channel, Command_Line.Argument(3));
      Status := Boolean'Input(Channel);
      Response := new String ' (String'Input(Channel));
      Print(Response.all, 0.02, False);
      String'Output(Channel, "Global Thermonuclear War");
      Status := Boolean'Input(Channel);
      if not Status then
         Command_Line.Set_Exit_Status(Command_Line.Failure);
         return;
      end if;
      Logname := new String ' (String'Input(Channel));
   else

      while not Status loop

         Response := new String ' (String'Input(Channel));
         Print(Response.all, 0.02, False);
         Command := new String ' (Text_Io.Get_Line);
         String'Output(Channel, Command.all);
         Status := Boolean'Input(Channel);
      end loop;
      Logname := new String ' (String'Input(Channel));
   end if;
   if Status then

      Gtk.Main.Init;

      Initialize(Xborg, channel, Logname);
      Gtk.Main.Main;
      Text_Io.Put("bye!");
      --Close_Socket(Socket);

   end if;
   if Xborg.Success then
      if Reborg_Status.End_Of_Program then
         Reborg_Status.End_Of_Program := False;

         Reborg_Parameters := Os_Lib.Argument_String_To_List("");
         Os_Lib.Normalize_Arguments(Reborg_Parameters.all);
         reborg_Id := Os_Lib.Non_Blocking_Spawn(Program_Name => Reborg_Location,
                                                Args => Reborg_Parameters.all);
      end if;
   end if;
exception

   when Args_Error =>
      Usage;
   when others =>
      if Xborg.Success then
         if Reborg_Status.End_Of_Program then
            Reborg_Status.End_Of_Program := False;

            Reborg_Parameters := Os_Lib.Argument_String_To_List("");
            Os_Lib.Normalize_Arguments(Reborg_Parameters.all);
            reborg_Id := Os_Lib.Non_Blocking_Spawn(Program_Name => Reborg_Location,
                                                   Args => Reborg_Parameters.all);
         end if;
      end if;
end Xborg;
